package ca.ubc.stat.blang.jvmmodel;

import blang.core.ConstantSupplier;
import blang.core.DeboxedName;
import blang.core.Distribution;
import blang.core.DistributionAdaptor;
import blang.core.ForwardSimulator;
import blang.core.IntDistribution;
import blang.core.IntDistributionAdaptor;
import blang.core.IntVar;
import blang.core.Model;
import blang.core.ModelBuilder;
import blang.core.ModelComponent;
import blang.core.Param;
import blang.core.RealDistribution;
import blang.core.RealDistributionAdaptor;
import blang.core.RealVar;
import blang.core.UnivariateModel;
import blang.core.WritableIntVar;
import blang.core.WritableRealVar;
import blang.inits.Arg;
import blang.inits.DesignatedConstructor;
import ca.ubc.stat.blang.StaticJavaUtils;
import ca.ubc.stat.blang.StaticUtils;
import ca.ubc.stat.blang.blangDsl.BlangDist;
import ca.ubc.stat.blang.blangDsl.BlangModel;
import ca.ubc.stat.blang.blangDsl.Dependency;
import ca.ubc.stat.blang.blangDsl.DistributionTypeSpecification;
import ca.ubc.stat.blang.blangDsl.ForLoop;
import ca.ubc.stat.blang.blangDsl.IndicatorDeclaration;
import ca.ubc.stat.blang.blangDsl.InitializerDependency;
import ca.ubc.stat.blang.blangDsl.InstantiatedDistribution;
import ca.ubc.stat.blang.blangDsl.JavaDist;
import ca.ubc.stat.blang.blangDsl.LawNode;
import ca.ubc.stat.blang.blangDsl.LogScaleFactorDeclaration;
import ca.ubc.stat.blang.blangDsl.VariableDeclaration;
import ca.ubc.stat.blang.blangDsl.VariableDeclarationComponent;
import ca.ubc.stat.blang.blangDsl.VariableType;
import ca.ubc.stat.blang.jvmmodel.BlangScope;
import ca.ubc.stat.blang.jvmmodel.XExpressionProcessor;
import com.google.common.base.Objects;
import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Parameter;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Random;
import java.util.function.Supplier;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtend.lib.annotations.Data;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtend2.lib.StringConcatenationClient;
import org.eclipse.xtext.common.types.JvmAnnotationReference;
import org.eclipse.xtext.common.types.JvmConstructor;
import org.eclipse.xtext.common.types.JvmDeclaredType;
import org.eclipse.xtext.common.types.JvmField;
import org.eclipse.xtext.common.types.JvmFormalParameter;
import org.eclipse.xtext.common.types.JvmGenericType;
import org.eclipse.xtext.common.types.JvmMember;
import org.eclipse.xtext.common.types.JvmOperation;
import org.eclipse.xtext.common.types.JvmTypeReference;
import org.eclipse.xtext.common.types.JvmVisibility;
import org.eclipse.xtext.naming.IQualifiedNameConverter;
import org.eclipse.xtext.resource.IResourceDescriptionsProvider;
import org.eclipse.xtext.xbase.XExpression;
import org.eclipse.xtext.xbase.annotations.xAnnotations.XAnnotation;
import org.eclipse.xtext.xbase.jvmmodel.JvmAnnotationReferenceBuilder;
import org.eclipse.xtext.xbase.jvmmodel.JvmTypeReferenceBuilder;
import org.eclipse.xtext.xbase.jvmmodel.JvmTypesBuilder;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.ExclusiveRange;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.lib.Pure;
import org.eclipse.xtext.xbase.lib.util.ToStringBuilder;

/**
 * SingleBlangModelInferrer gets instantiated for each model being inferred.
 */
@SuppressWarnings("all")
public class SingleBlangModelInferrer {
  @Data
  public static class ConstructorArgument {
    private final JvmTypeReference deboxedType;
    
    private final boolean isParam;
    
    public ConstructorArgument(final JvmTypeReference deboxedType, final boolean isParam) {
      super();
      this.deboxedType = deboxedType;
      this.isParam = isParam;
    }
    
    @Override
    @Pure
    public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result + ((this.deboxedType== null) ? 0 : this.deboxedType.hashCode());
      result = prime * result + (this.isParam ? 1231 : 1237);
      return result;
    }
    
    @Override
    @Pure
    public boolean equals(final Object obj) {
      if (this == obj)
        return true;
      if (obj == null)
        return false;
      if (getClass() != obj.getClass())
        return false;
      SingleBlangModelInferrer.ConstructorArgument other = (SingleBlangModelInferrer.ConstructorArgument) obj;
      if (this.deboxedType == null) {
        if (other.deboxedType != null)
          return false;
      } else if (!this.deboxedType.equals(other.deboxedType))
        return false;
      if (other.isParam != this.isParam)
        return false;
      return true;
    }
    
    @Override
    @Pure
    public String toString() {
      ToStringBuilder b = new ToStringBuilder(this);
      b.add("deboxedType", this.deboxedType);
      b.add("isParam", this.isParam);
      return b.toString();
    }
    
    @Pure
    public JvmTypeReference getDeboxedType() {
      return this.deboxedType;
    }
    
    @Pure
    public boolean isParam() {
      return this.isParam;
    }
  }
  
  private final BlangModel model;
  
  private final JvmDeclaredType output;
  
  private final XExpressionProcessor xExpressions;
  
  @Extension
  private final JvmTypesBuilder _typeBuilder;
  
  @Extension
  private final JvmAnnotationReferenceBuilder _annotationTypesBuilder;
  
  @Extension
  private final JvmTypeReferenceBuilder _typeReferenceBuilder;
  
  @Extension
  private final IResourceDescriptionsProvider index;
  
  private final IQualifiedNameConverter qualNameConverter;
  
  public void infer() {
    this.setupClass();
    final JvmGenericType builderOutput = this.setupBuilder();
    final BlangScope globalScope = this.setupVariables(builderOutput);
    this.setupMain(builderOutput);
    StaticUtils.eagerlyEvaluate(this.componentsMethodBody(globalScope));
    StaticUtils.eagerlyEvaluate(this.builderBody(globalScope));
    boolean _hasForwardGenerationBlock = this.hasForwardGenerationBlock();
    if (_hasForwardGenerationBlock) {
      StaticUtils.eagerlyEvaluate(this.forwardSimulationMethodBody(globalScope));
    }
    this.xExpressions.endAuxiliaryMethodGenerationPhase();
    this.generateConstructor(globalScope);
    this.generateBuilder(globalScope, builderOutput);
    this.generateComponentsMethod(globalScope);
    this.generateForwardSimulationMethod(globalScope);
    boolean _isPresent = this.uniqueRandomVariable().isPresent();
    if (_isPresent) {
      this.generateGetRealization();
      this.generateStaticDistributionBuilder(globalScope);
    }
  }
  
  public boolean setupMain(final JvmGenericType type) {
    boolean _xblockexpression = false;
    {
      final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
        StringConcatenation _builder = new StringConcatenation();
        _builder.append("Utility main method for posterior inference on this model");
        this._typeBuilder.setDocumentation(it, _builder.toString());
        final String paramName = "arguments";
        EList<JvmFormalParameter> _parameters = it.getParameters();
        JvmFormalParameter _parameter = this._typeBuilder.toParameter(this.model, paramName, this._typeReferenceBuilder.typeRef(StaticJavaUtils.STRING_ARRAY));
        this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(StaticJavaUtils.class);
            _builder.append(_typeRef);
            _builder.append(".callRunner(");
            _builder.append(SingleBlangModelInferrer.BUILDER_NAME);
            _builder.append(".class, ");
            _builder.append(paramName);
            _builder.append(");");
            _builder.newLineIfNotEmpty();
          }
        };
        this._typeBuilder.setBody(it, _client);
      };
      final JvmOperation main = this._typeBuilder.toMethod(this.model, "main", this._typeReferenceBuilder.typeRef(Void.TYPE), _function);
      main.setStatic(true);
      EList<JvmMember> _members = this.output.getMembers();
      _xblockexpression = this._typeBuilder.<JvmOperation>operator_add(_members, main);
    }
    return _xblockexpression;
  }
  
  private void setupClass() {
    String _package = this.model.getPackage();
    boolean _tripleNotEquals = (_package != null);
    if (_tripleNotEquals) {
      this.output.setPackageName(this.model.getPackage());
    }
    EList<XAnnotation> _annotations = this.model.getAnnotations();
    for (final XAnnotation annotation : _annotations) {
      this._typeBuilder.addAnnotation(this.output, annotation);
    }
    EList<JvmTypeReference> _superTypes = this.output.getSuperTypes();
    JvmTypeReference _typeRef = this._typeReferenceBuilder.typeRef(Model.class);
    this._typeBuilder.<JvmTypeReference>operator_add(_superTypes, _typeRef);
    boolean _isPresent = this.uniqueRandomVariable().isPresent();
    if (_isPresent) {
      EList<JvmTypeReference> _superTypes_1 = this.output.getSuperTypes();
      JvmTypeReference _typeRef_1 = this._typeReferenceBuilder.typeRef(UnivariateModel.class, this.uniqueRandomVariable().get().deboxedType);
      this._typeBuilder.<JvmTypeReference>operator_add(_superTypes_1, _typeRef_1);
    }
    boolean _hasForwardGenerationBlock = this.hasForwardGenerationBlock();
    if (_hasForwardGenerationBlock) {
      EList<JvmTypeReference> _superTypes_2 = this.output.getSuperTypes();
      JvmTypeReference _typeRef_2 = this._typeReferenceBuilder.typeRef(ForwardSimulator.class);
      this._typeBuilder.<JvmTypeReference>operator_add(_superTypes_2, _typeRef_2);
    }
  }
  
  private boolean hasForwardGenerationBlock() {
    XExpression _generationAlgorithm = this.model.getGenerationAlgorithm();
    return (_generationAlgorithm != null);
  }
  
  private JvmGenericType setupBuilder() {
    final Procedure1<JvmGenericType> _function = (JvmGenericType it) -> {
      it.setStatic(true);
    };
    final JvmGenericType builderOutput = this._typeBuilder.toClass(this.model, SingleBlangModelInferrer.BUILDER_NAME, _function);
    EList<JvmMember> _members = builderOutput.getMembers();
    final Procedure1<JvmField> _function_1 = (JvmField it) -> {
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("false");
        }
      };
      this._typeBuilder.setInitializer(it, _client);
    };
    JvmField _field = this._typeBuilder.toField(this.model, SingleBlangModelInferrer.IS_FROM_CMD_LINE_ARG, this._typeReferenceBuilder.typeRef("boolean"), _function_1);
    this._typeBuilder.<JvmField>operator_add(_members, _field);
    EList<JvmTypeReference> _superTypes = builderOutput.getSuperTypes();
    JvmTypeReference _typeRef = this._typeReferenceBuilder.typeRef(ModelBuilder.class);
    this._typeBuilder.<JvmTypeReference>operator_add(_superTypes, _typeRef);
    EList<JvmMember> _members_1 = this.output.getMembers();
    this._typeBuilder.<JvmGenericType>operator_add(_members_1, builderOutput);
    EList<JvmMember> _members_2 = this.output.getMembers();
    final Procedure1<JvmOperation> _function_2 = (JvmOperation it) -> {
      it.setStatic(true);
      EList<JvmAnnotationReference> _annotations = it.getAnnotations();
      JvmAnnotationReference _annotationRef = this._annotationTypesBuilder.annotationRef(DesignatedConstructor.class);
      this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append(SingleBlangModelInferrer.BUILDER_NAME);
          _builder.append(" result = new ");
          _builder.append(SingleBlangModelInferrer.BUILDER_NAME);
          _builder.append("();");
          _builder.newLineIfNotEmpty();
          _builder.append("result.");
          _builder.append(SingleBlangModelInferrer.IS_FROM_CMD_LINE_ARG);
          _builder.append(" = true;");
          _builder.newLineIfNotEmpty();
          _builder.append("return result;");
          _builder.newLine();
        }
      };
      this._typeBuilder.setBody(it, _client);
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, "builderFromCommandLine", this._typeReferenceBuilder.typeRef(builderOutput), _function_2);
    this._typeBuilder.<JvmOperation>operator_add(_members_2, _method);
    return builderOutput;
  }
  
  private Optional<BlangScope.BlangVariable> uniqueRandomVariable() {
    Optional<BlangScope.BlangVariable> result = Optional.<BlangScope.BlangVariable>empty();
    EList<VariableDeclaration> _variableDeclarations = this.model.getVariableDeclarations();
    for (final VariableDeclaration variableDeclaration : _variableDeclarations) {
      VariableType _variableType = variableDeclaration.getVariableType();
      boolean _equals = Objects.equal(_variableType, VariableType.RANDOM);
      if (_equals) {
        EList<VariableDeclarationComponent> _components = variableDeclaration.getComponents();
        for (final VariableDeclarationComponent varDeclComponent : _components) {
          boolean _isPresent = result.isPresent();
          if (_isPresent) {
            return Optional.<BlangScope.BlangVariable>empty();
          } else {
            JvmTypeReference _type = variableDeclaration.getType();
            String _varName = this.varName(varDeclComponent);
            BlangScope.BlangVariable _blangVariable = new BlangScope.BlangVariable(_type, _varName, false);
            result = Optional.<BlangScope.BlangVariable>of(_blangVariable);
          }
        }
      }
    }
    return result;
  }
  
  private String isInitializedMethodName(final String deboxedName) {
    return (deboxedName + "_initialized");
  }
  
  private BlangScope setupVariables(final JvmGenericType builderOutput) {
    final BlangScope result = BlangScope.emptyScope();
    EList<VariableDeclaration> _variableDeclarations = this.model.getVariableDeclarations();
    for (final VariableDeclaration variableDeclaration : _variableDeclarations) {
      EList<VariableDeclarationComponent> _components = variableDeclaration.getComponents();
      for (final VariableDeclarationComponent varDeclComponent : _components) {
        {
          final String name = this.varName(varDeclComponent);
          JvmTypeReference _type = variableDeclaration.getType();
          boolean _isParam = StaticUtils.isParam(variableDeclaration.getVariableType());
          final BlangScope.BlangVariable blangVariable = new BlangScope.BlangVariable(_type, name, _isParam);
          result.operator_add(blangVariable);
          EList<JvmMember> _members = this.output.getMembers();
          final Procedure1<JvmField> _function = (JvmField it) -> {
            EList<XAnnotation> _annotations = variableDeclaration.getAnnotations();
            for (final XAnnotation annotation : _annotations) {
              this._typeBuilder.addAnnotation(it, annotation);
            }
            it.setFinal(true);
            boolean _isParam_1 = blangVariable.isParam();
            if (_isParam_1) {
              EList<JvmAnnotationReference> _annotations_1 = it.getAnnotations();
              JvmAnnotationReference _annotationRef = this._annotationTypesBuilder.annotationRef(Param.class);
              this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations_1, _annotationRef);
            }
          };
          JvmField _field = this._typeBuilder.toField(variableDeclaration, blangVariable.boxedName(), blangVariable.boxedType(this._typeReferenceBuilder), _function);
          this._typeBuilder.<JvmField>operator_add(_members, _field);
          EList<JvmMember> _members_1 = this.output.getMembers();
          final Procedure1<JvmOperation> _function_1 = (JvmOperation it) -> {
            StringConcatenationClient _client = new StringConcatenationClient() {
              @Override
              protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                _builder.append("return ");
                String _deboxingInvocationString = blangVariable.deboxingInvocationString();
                _builder.append(_deboxingInvocationString);
                _builder.append(";");
                _builder.newLineIfNotEmpty();
              }
            };
            this._typeBuilder.setBody(it, _client);
          };
          JvmOperation _method = this._typeBuilder.toMethod(variableDeclaration, StaticUtils.getterName(blangVariable.deboxedName), blangVariable.deboxedType, _function_1);
          this._typeBuilder.<JvmOperation>operator_add(_members_1, _method);
          XExpression _varInitBlock = varDeclComponent.getVarInitBlock();
          final boolean optional = (_varInitBlock != null);
          EList<JvmMember> _members_2 = builderOutput.getMembers();
          final Procedure1<JvmField> _function_2 = (JvmField it) -> {
            it.setVisibility(JvmVisibility.PUBLIC);
            EList<JvmAnnotationReference> _annotations = it.getAnnotations();
            JvmAnnotationReference _annotationRef = this._annotationTypesBuilder.annotationRef(Arg.class);
            this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          };
          JvmField _field_1 = this._typeBuilder.toField(variableDeclaration, blangVariable.deboxedName, this.optionalize(blangVariable.deboxedType, optional), _function_2);
          this._typeBuilder.<JvmField>operator_add(_members_2, _field_1);
          if ((!optional)) {
            EList<JvmMember> _members_3 = builderOutput.getMembers();
            final Procedure1<JvmField> _function_3 = (JvmField it) -> {
              it.setVisibility(JvmVisibility.PRIVATE);
              StringConcatenationClient _client = new StringConcatenationClient() {
                @Override
                protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                  _builder.append("false");
                }
              };
              this._typeBuilder.setInitializer(it, _client);
            };
            JvmField _field_2 = this._typeBuilder.toField(variableDeclaration, this.isInitializedMethodName(blangVariable.deboxedName), this._typeReferenceBuilder.typeRef("boolean"), _function_3);
            this._typeBuilder.<JvmField>operator_add(_members_3, _field_2);
          }
          EList<JvmMember> _members_4 = builderOutput.getMembers();
          final Procedure1<JvmOperation> _function_4 = (JvmOperation it) -> {
            EList<JvmFormalParameter> _parameters = it.getParameters();
            JvmFormalParameter _parameter = this._typeBuilder.toParameter(this.model, blangVariable.deboxedName, blangVariable.deboxedType);
            this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
            it.setVisibility(JvmVisibility.PUBLIC);
            StringConcatenationClient _client = new StringConcatenationClient() {
              @Override
              protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
                {
                  if (optional) {
                    _builder.append("// work around typeRef(..) limitation");
                    _builder.newLine();
                    JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(Optional.class, blangVariable.deboxedType);
                    _builder.append(_typeRef);
                    _builder.append(" ");
                    String _generatedName = StaticUtils.generatedName("dummy");
                    _builder.append(_generatedName);
                    _builder.append(" = null;");
                    _builder.newLineIfNotEmpty();
                  } else {
                    String _isInitializedMethodName = SingleBlangModelInferrer.this.isInitializedMethodName(blangVariable.deboxedName);
                    _builder.append(_isInitializedMethodName);
                    _builder.append(" = true;");
                    _builder.newLineIfNotEmpty();
                  }
                }
                _builder.append("this.");
                _builder.append(blangVariable.deboxedName);
                _builder.append(" = ");
                {
                  if (optional) {
                    _builder.append("Optional.of(");
                    _builder.append(blangVariable.deboxedName);
                    _builder.append(")");
                  } else {
                    _builder.append(blangVariable.deboxedName);
                  }
                }
                _builder.append(";");
                _builder.newLineIfNotEmpty();
                _builder.append("return this;");
                _builder.newLine();
              }
            };
            this._typeBuilder.setBody(it, _client);
          };
          JvmOperation _method_1 = this._typeBuilder.toMethod(variableDeclaration, StaticUtils.setterName(blangVariable.deboxedName), this._typeReferenceBuilder.typeRef(builderOutput), _function_4);
          this._typeBuilder.<JvmOperation>operator_add(_members_4, _method_1);
        }
      }
    }
    return result;
  }
  
  public JvmTypeReference optionalize(final JvmTypeReference deboxedType, final boolean makeOptional) {
    if (makeOptional) {
      return this._typeReferenceBuilder.typeRef(Optional.class, deboxedType);
    } else {
      return deboxedType;
    }
  }
  
  private static List<BlangScope.BlangVariable> constructorOrder(final BlangScope scope) {
    List<BlangScope.BlangVariable> result = new ArrayList<BlangScope.BlangVariable>();
    final boolean[] variablesOrder = { false, true };
    for (final boolean isParam : variablesOrder) {
      final Function1<BlangScope.BlangVariable, Boolean> _function = (BlangScope.BlangVariable it) -> {
        boolean _isParam = it.isParam();
        return Boolean.valueOf((Boolean.valueOf(_isParam) == Boolean.valueOf(isParam)));
      };
      Iterable<BlangScope.BlangVariable> _filter = IterableExtensions.<BlangScope.BlangVariable>filter(scope.variables(), _function);
      for (final BlangScope.BlangVariable variable : _filter) {
        result.add(variable);
      }
    }
    return result;
  }
  
  private void generateBuilder(final BlangScope scope, final JvmGenericType builderOutput) {
    EList<JvmMember> _members = builderOutput.getMembers();
    final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      this._typeBuilder.setBody(it, this.builderBody(scope));
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, SingleBlangModelInferrer.BUILDER_METHOD_NAME, this._typeReferenceBuilder.typeRef(this.output), _function);
    this._typeBuilder.<JvmOperation>operator_add(_members, _method);
  }
  
  public String varName(final VariableDeclarationComponent component) {
    return component.getName().getName();
  }
  
  public String varName(final ForLoop component) {
    return component.getName().getName();
  }
  
  private StringConcatenationClient builderBody(final BlangScope globalScope) {
    final String prefixForFinalVariable = "__";
    final BlangScope incrementalScope = BlangScope.emptyScope();
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append("// For each optional type, either get the value, or evaluate the ?: expression");
        _builder.newLine();
        {
          EList<VariableDeclaration> _variableDeclarations = SingleBlangModelInferrer.this.model.getVariableDeclarations();
          for(final VariableDeclaration variableDeclaration : _variableDeclarations) {
            {
              EList<VariableDeclarationComponent> _components = variableDeclaration.getComponents();
              for(final VariableDeclarationComponent varDeclComponent : _components) {
                {
                  XExpression _varInitBlock = varDeclComponent.getVarInitBlock();
                  boolean _tripleNotEquals = (_varInitBlock != null);
                  if (_tripleNotEquals) {
                    JvmTypeReference _type = variableDeclaration.getType();
                    _builder.append(_type);
                    _builder.append(" ");
                    String _varName = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName);
                    _builder.append(";");
                    _builder.newLineIfNotEmpty();
                    _builder.append("if (this.");
                    String _varName_1 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_1);
                    _builder.append(" != null && this.");
                    String _varName_2 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_2);
                    _builder.append(".isPresent()) {");
                    _builder.newLineIfNotEmpty();
                    _builder.append("  ");
                    String _varName_3 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_3, "  ");
                    _builder.append(" = this.");
                    String _varName_4 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_4, "  ");
                    _builder.append(".get();");
                    _builder.newLineIfNotEmpty();
                    _builder.append("} else {");
                    _builder.newLine();
                    _builder.append("  ");
                    String _varName_5 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_5, "  ");
                    _builder.append(" = ");
                    StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(varDeclComponent.getVarInitBlock(), incrementalScope, variableDeclaration.getType());
                    _builder.append(_process, "  ");
                    _builder.append(";");
                    _builder.newLineIfNotEmpty();
                    _builder.append("}");
                    _builder.newLine();
                  } else {
                    _builder.append("if (!");
                    _builder.append(SingleBlangModelInferrer.IS_FROM_CMD_LINE_ARG);
                    _builder.append(" && !");
                    String _isInitializedMethodName = SingleBlangModelInferrer.this.isInitializedMethodName(SingleBlangModelInferrer.this.varName(varDeclComponent));
                    _builder.append(_isInitializedMethodName);
                    _builder.append(")");
                    _builder.newLineIfNotEmpty();
                    _builder.append("  ");
                    _builder.append("throw new RuntimeException(\"Not all fields were set in the builder, e.g. missing ");
                    String _varName_6 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                    _builder.append(_varName_6, "  ");
                    _builder.append("\");");
                    _builder.newLineIfNotEmpty();
                  }
                }
                _builder.append("final ");
                JvmTypeReference _type_1 = variableDeclaration.getType();
                _builder.append(_type_1);
                _builder.append(" ");
                _builder.append(prefixForFinalVariable);
                String _varName_7 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                _builder.append(_varName_7);
                _builder.append(" = ");
                String _varName_8 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                _builder.append(_varName_8);
                _builder.append(";");
                _builder.newLineIfNotEmpty();
                JvmTypeReference _type_2 = variableDeclaration.getType();
                String _varName_9 = SingleBlangModelInferrer.this.varName(varDeclComponent);
                BlangScope.BlangVariable _blangVariable = new BlangScope.BlangVariable(_type_2, _varName_9, false);
                BlangScope.BlangVariable _add = incrementalScope.operator_add(_blangVariable);
                _builder.append(_add);
                _builder.newLineIfNotEmpty();
              }
            }
          }
        }
        _builder.append("// Build the instance after boxing params");
        _builder.newLine();
        _builder.append("return new ");
        JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(SingleBlangModelInferrer.this.output);
        _builder.append(_typeRef);
        _builder.append("(");
        _builder.newLineIfNotEmpty();
        {
          List<BlangScope.BlangVariable> _constructorOrder = SingleBlangModelInferrer.constructorOrder(globalScope);
          boolean _hasElements = false;
          for(final BlangScope.BlangVariable variable : _constructorOrder) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(", ", "  ");
            }
            {
              boolean _isParam = variable.isParam();
              if (_isParam) {
                _builder.append("  ");
                _builder.append("new ");
                JvmTypeReference _typeRef_1 = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(ConstantSupplier.class);
                _builder.append(_typeRef_1, "  ");
                _builder.append("(");
                _builder.append(prefixForFinalVariable, "  ");
                _builder.append(variable.deboxedName, "  ");
                _builder.append(")");
                _builder.newLineIfNotEmpty();
              } else {
                _builder.append("  ");
                _builder.append(prefixForFinalVariable, "  ");
                _builder.append(variable.deboxedName, "  ");
                _builder.newLineIfNotEmpty();
              }
            }
          }
        }
        _builder.append(");");
        _builder.newLine();
      }
    };
    return _client;
  }
  
  private void generateGetRealization() {
    EList<JvmMember> _members = this.output.getMembers();
    final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("return ");
          _builder.append(SingleBlangModelInferrer.this.uniqueRandomVariable().get().deboxedName);
          _builder.append(";");
          _builder.newLineIfNotEmpty();
        }
      };
      this._typeBuilder.setBody(it, _client);
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, StaticUtils.uniqueDeclaredMethod(UnivariateModel.class), this.uniqueRandomVariable().get().deboxedType, _function);
    this._typeBuilder.<JvmOperation>operator_add(_members, _method);
  }
  
  private void generateStaticDistributionBuilder(final BlangScope scope) {
    EList<JvmMember> _members = this.output.getMembers();
    final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      it.setStatic(true);
      List<BlangScope.BlangVariable> _constructorOrder = SingleBlangModelInferrer.constructorOrder(scope);
      for (final BlangScope.BlangVariable variable : _constructorOrder) {
        if (((!this.isRealOrIntDistributionType()) || variable.isParam())) {
          final JvmFormalParameter param = this._typeBuilder.toParameter(this.model, variable.deboxedName, variable.deboxedType);
          boolean _isParam = variable.isParam();
          if (_isParam) {
            EList<JvmAnnotationReference> _annotations = param.getAnnotations();
            JvmAnnotationReference _annotationRef = this._annotationTypesBuilder.annotationRef(Param.class);
            this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          }
          EList<JvmFormalParameter> _parameters = it.getParameters();
          this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, param);
        }
      }
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(UnivariateModel.class, SingleBlangModelInferrer.this.uniqueRandomVariable().get().deboxedType);
          _builder.append(_typeRef);
          _builder.append(" univariateModel = new ");
          String _name = SingleBlangModelInferrer.this.model.getName();
          _builder.append(_name);
          _builder.append("(");
          _builder.newLineIfNotEmpty();
          {
            List<BlangScope.BlangVariable> _constructorOrder = SingleBlangModelInferrer.constructorOrder(scope);
            boolean _hasElements = false;
            for(final BlangScope.BlangVariable variable : _constructorOrder) {
              if (!_hasElements) {
                _hasElements = true;
              } else {
                _builder.appendImmediate(", ", "  ");
              }
              {
                boolean _isParam = variable.isParam();
                if (_isParam) {
                  _builder.append("  ");
                  _builder.append("new ");
                  JvmTypeReference _typeRef_1 = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(ConstantSupplier.class);
                  _builder.append(_typeRef_1, "  ");
                  _builder.append("(");
                  _builder.append(variable.deboxedName, "  ");
                  _builder.append(")");
                  _builder.newLineIfNotEmpty();
                } else {
                  {
                    boolean _isRealOrIntDistributionType = SingleBlangModelInferrer.this.isRealOrIntDistributionType();
                    if (_isRealOrIntDistributionType) {
                      _builder.append("  ");
                      _builder.append("new ");
                      JvmTypeReference _writableVariableImpl = SingleBlangModelInferrer.this.writableVariableImpl();
                      _builder.append(_writableVariableImpl, "  ");
                      _builder.append("()");
                      _builder.newLineIfNotEmpty();
                    } else {
                      _builder.append("  ");
                      _builder.append(variable.deboxedName, "  ");
                      _builder.newLineIfNotEmpty();
                    }
                  }
                }
              }
            }
          }
          _builder.append(");");
          _builder.newLine();
          JvmTypeReference _typeRef_2 = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(Distribution.class, SingleBlangModelInferrer.this.uniqueRandomVariable().get().deboxedType);
          _builder.append(_typeRef_2);
          _builder.append(" distribution = new ");
          JvmTypeReference _typeRef_3 = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(DistributionAdaptor.class);
          _builder.append(_typeRef_3);
          _builder.append("(univariateModel);");
          _builder.newLineIfNotEmpty();
          {
            boolean _isRealOrIntDistributionType_1 = SingleBlangModelInferrer.this.isRealOrIntDistributionType();
            if (_isRealOrIntDistributionType_1) {
              _builder.append("return new ");
              JvmTypeReference _adaptorType = SingleBlangModelInferrer.this.adaptorType();
              _builder.append(_adaptorType);
              _builder.append("(distribution);");
              _builder.newLineIfNotEmpty();
            } else {
              _builder.append("return distribution;");
              _builder.newLine();
            }
          }
        }
      };
      this._typeBuilder.setBody(it, _client);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("Returns an instance with fixed parameters values and conforming the Distribution interface. ");
      _builder.newLine();
      _builder.append("Useful when passing around distributions as parameters, e.g. for Dirichlet Process mixtures. ");
      _builder.newLine();
      this._typeBuilder.setDocumentation(it, _builder.toString());
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, "distribution", this.distributionType(), _function);
    this._typeBuilder.<JvmOperation>operator_add(_members, _method);
  }
  
  public JvmTypeReference adaptorType() {
    JvmTypeReference _switchResult = null;
    String _qualifiedName = this.uniqueRandomVariable().get().deboxedType.getQualifiedName();
    boolean _matched = false;
    String _canonicalName = RealVar.class.getCanonicalName();
    if (Objects.equal(_qualifiedName, _canonicalName)) {
      _matched=true;
      _switchResult = this._typeReferenceBuilder.typeRef(RealDistributionAdaptor.class);
    }
    if (!_matched) {
      String _canonicalName_1 = IntVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName, _canonicalName_1)) {
        _matched=true;
        _switchResult = this._typeReferenceBuilder.typeRef(IntDistributionAdaptor.class);
      }
    }
    if (!_matched) {
      throw new RuntimeException();
    }
    return _switchResult;
  }
  
  public JvmTypeReference distributionType() {
    JvmTypeReference _switchResult = null;
    String _qualifiedName = this.uniqueRandomVariable().get().deboxedType.getQualifiedName();
    boolean _matched = false;
    String _canonicalName = RealVar.class.getCanonicalName();
    if (Objects.equal(_qualifiedName, _canonicalName)) {
      _matched=true;
      _switchResult = this._typeReferenceBuilder.typeRef(RealDistribution.class);
    }
    if (!_matched) {
      String _canonicalName_1 = IntVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName, _canonicalName_1)) {
        _matched=true;
        _switchResult = this._typeReferenceBuilder.typeRef(IntDistribution.class);
      }
    }
    if (!_matched) {
      _switchResult = this._typeReferenceBuilder.typeRef(Distribution.class, this.uniqueRandomVariable().get().deboxedType);
    }
    return _switchResult;
  }
  
  public JvmTypeReference writableVariableImpl() {
    JvmTypeReference _switchResult = null;
    String _qualifiedName = this.uniqueRandomVariable().get().deboxedType.getQualifiedName();
    boolean _matched = false;
    String _canonicalName = RealVar.class.getCanonicalName();
    if (Objects.equal(_qualifiedName, _canonicalName)) {
      _matched=true;
      _switchResult = this._typeReferenceBuilder.typeRef(RealDistributionAdaptor.WritableRealVarImpl.class);
    }
    if (!_matched) {
      String _canonicalName_1 = IntVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName, _canonicalName_1)) {
        _matched=true;
        _switchResult = this._typeReferenceBuilder.typeRef(IntDistributionAdaptor.WritableIntVarImpl.class);
      }
    }
    if (!_matched) {
      throw new RuntimeException();
    }
    return _switchResult;
  }
  
  public boolean isRealOrIntDistributionType() {
    boolean _switchResult = false;
    String _qualifiedName = this.uniqueRandomVariable().get().deboxedType.getQualifiedName();
    boolean _matched = false;
    String _canonicalName = RealVar.class.getCanonicalName();
    if (Objects.equal(_qualifiedName, _canonicalName)) {
      _matched=true;
      _switchResult = true;
    }
    if (!_matched) {
      String _canonicalName_1 = IntVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName, _canonicalName_1)) {
        _matched=true;
        _switchResult = true;
      }
    }
    if (!_matched) {
      _switchResult = false;
    }
    return _switchResult;
  }
  
  private void generateConstructor(final BlangScope scope) {
    EList<JvmMember> _members = this.output.getMembers();
    final Procedure1<JvmConstructor> _function = (JvmConstructor it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      List<BlangScope.BlangVariable> _constructorOrder = SingleBlangModelInferrer.constructorOrder(scope);
      for (final BlangScope.BlangVariable variable : _constructorOrder) {
        {
          final JvmFormalParameter param = this._typeBuilder.toParameter(this.model, variable.boxedName(), variable.boxedType(this._typeReferenceBuilder));
          boolean _isParam = variable.isParam();
          if (_isParam) {
            EList<JvmAnnotationReference> _annotations = param.getAnnotations();
            JvmAnnotationReference _annotationRef = this._annotationTypesBuilder.annotationRef(Param.class);
            this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations, _annotationRef);
          }
          EList<JvmAnnotationReference> _annotations_1 = param.getAnnotations();
          JvmAnnotationReference _annotationRef_1 = this._annotationTypesBuilder.annotationRef(DeboxedName.class, variable.deboxedName);
          this._typeBuilder.<JvmAnnotationReference>operator_add(_annotations_1, _annotationRef_1);
          EList<JvmFormalParameter> _parameters = it.getParameters();
          this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, param);
        }
      }
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          {
            Iterable<BlangScope.BlangVariable> _variables = scope.variables();
            for(final BlangScope.BlangVariable variable : _variables) {
              _builder.append("this.");
              String _boxedName = variable.boxedName();
              _builder.append(_boxedName);
              _builder.append(" = ");
              String _boxedName_1 = variable.boxedName();
              _builder.append(_boxedName_1);
              _builder.append(";");
              _builder.newLineIfNotEmpty();
            }
          }
        }
      };
      this._typeBuilder.setBody(it, _client);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("Note: the generated code has the following properties used at runtime:");
      _builder.newLine();
      _builder.append("  ");
      _builder.append("- all arguments are annotated with a BlangVariable annotation");
      _builder.newLine();
      _builder.append("  ");
      _builder.append("- params additionally have a Param annotation");
      _builder.newLine();
      _builder.append("  ");
      _builder.append("- the order of the arguments is as follows:");
      _builder.newLine();
      _builder.append("    ");
      _builder.append("- first, all the random variables in the order they occur in the blang file");
      _builder.newLine();
      _builder.append("    ");
      _builder.append("- second, all the params in the order they occur in the blang file");
      _builder.newLine();
      this._typeBuilder.setDocumentation(it, _builder.toString());
    };
    JvmConstructor _constructor = this._typeBuilder.toConstructor(this.model, _function);
    this._typeBuilder.<JvmConstructor>operator_add(_members, _constructor);
  }
  
  private void generateForwardSimulationMethod(final BlangScope scope) {
    boolean _hasForwardGenerationBlock = this.hasForwardGenerationBlock();
    boolean _not = (!_hasForwardGenerationBlock);
    if (_not) {
      return;
    }
    EList<JvmMember> _members = this.output.getMembers();
    final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      EList<JvmFormalParameter> _parameters = it.getParameters();
      JvmFormalParameter _parameter = this._typeBuilder.toParameter(this.model.getGenerationAlgorithm(), this.model.getGenerationRandom(), this._typeReferenceBuilder.typeRef(Random.class));
      this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
      this._typeBuilder.setBody(it, this.forwardSimulationMethodBody(scope));
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, SingleBlangModelInferrer.FWD_SIM_METHOD_NAME, this._typeReferenceBuilder.typeRef(Void.TYPE), _function);
    this._typeBuilder.<JvmOperation>operator_add(_members, _method);
  }
  
  private StringConcatenationClient forwardSimulationMethodBody(final BlangScope scope) {
    final Optional<BlangScope.BlangVariable> uniqueRandomVariable = this.uniqueRandomVariable();
    Class<? extends Number> _xifexpression = null;
    boolean _isPresent = uniqueRandomVariable.isPresent();
    if (_isPresent) {
      Class<? extends Number> _switchResult = null;
      String _qualifiedName = uniqueRandomVariable.get().deboxedType.getQualifiedName();
      boolean _matched = false;
      String _canonicalName = RealVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName, _canonicalName)) {
        _matched=true;
        _switchResult = Double.class;
      }
      if (!_matched) {
        String _canonicalName_1 = IntVar.class.getCanonicalName();
        if (Objects.equal(_qualifiedName, _canonicalName_1)) {
          _matched=true;
          _switchResult = Integer.class;
        }
      }
      if (!_matched) {
        _switchResult = null;
      }
      _xifexpression = _switchResult;
    } else {
      _xifexpression = null;
    }
    final Class<? extends Number> deboxedType = _xifexpression;
    Class<?> _xifexpression_1 = null;
    boolean _isPresent_1 = uniqueRandomVariable.isPresent();
    if (_isPresent_1) {
      Class<?> _switchResult_1 = null;
      String _qualifiedName_1 = uniqueRandomVariable.get().deboxedType.getQualifiedName();
      boolean _matched_1 = false;
      String _canonicalName_2 = RealVar.class.getCanonicalName();
      if (Objects.equal(_qualifiedName_1, _canonicalName_2)) {
        _matched_1=true;
        _switchResult_1 = WritableRealVar.class;
      }
      if (!_matched_1) {
        String _canonicalName_3 = IntVar.class.getCanonicalName();
        if (Objects.equal(_qualifiedName_1, _canonicalName_3)) {
          _matched_1=true;
          _switchResult_1 = WritableIntVar.class;
        }
      }
      if (!_matched_1) {
        _switchResult_1 = null;
      }
      _xifexpression_1 = _switchResult_1;
    } else {
      _xifexpression_1 = null;
    }
    final Class<?> writableType = _xifexpression_1;
    JvmTypeReference _typeRef = this._typeReferenceBuilder.typeRef(Random.class);
    String _generationRandom = this.model.getGenerationRandom();
    BlangScope.BlangVariable _blangVariable = new BlangScope.BlangVariable(_typeRef, _generationRandom, false);
    final BlangScope scopeWithRandom = scope.child(_blangVariable);
    StringConcatenationClient _xifexpression_2 = null;
    if ((uniqueRandomVariable.isPresent() && (deboxedType != null))) {
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("((");
          JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(writableType);
          _builder.append(_typeRef);
          _builder.append(") ");
          _builder.append(uniqueRandomVariable.get().deboxedName);
          _builder.append(").set(");
          _builder.newLineIfNotEmpty();
          _builder.append("  ");
          StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(SingleBlangModelInferrer.this.model.getGenerationAlgorithm(), scopeWithRandom, SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(deboxedType));
          _builder.append(_process, "  ");
          _builder.newLineIfNotEmpty();
          _builder.append(");");
          _builder.newLine();
        }
      };
      _xifexpression_2 = _client;
    } else {
      StringConcatenationClient _client_1 = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(SingleBlangModelInferrer.this.model.getGenerationAlgorithm(), scopeWithRandom, SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(Void.TYPE));
          _builder.append(_process);
          _builder.append(";");
          _builder.newLineIfNotEmpty();
        }
      };
      _xifexpression_2 = _client_1;
    }
    return _xifexpression_2;
  }
  
  private void generateComponentsMethod(final BlangScope scope) {
    EList<JvmMember> _members = this.output.getMembers();
    final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
      it.setVisibility(JvmVisibility.PUBLIC);
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("A component can be either a distribution, support constraint, or another model  ");
      _builder.newLine();
      _builder.append("which recursively defines additional components.");
      this._typeBuilder.setDocumentation(it, _builder.toString());
      this._typeBuilder.setBody(it, this.componentsMethodBody(scope));
    };
    JvmOperation _method = this._typeBuilder.toMethod(this.model, SingleBlangModelInferrer.COMPONENTS_METHOD_NAME, this._typeReferenceBuilder.typeRef(Collection.class, this._typeReferenceBuilder.typeRef(ModelComponent.class)), _function);
    this._typeBuilder.<JvmOperation>operator_add(_members, _method);
  }
  
  /**
   * Entry point of generation of the body of the method components(), the core of a probabilistic model.
   */
  private StringConcatenationClient componentsMethodBody(final BlangScope scope) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        JvmTypeReference _typeRef = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(ArrayList.class, SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(ModelComponent.class));
        _builder.append(_typeRef);
        _builder.append(" ");
        _builder.append(SingleBlangModelInferrer.COMPONENTS_LIST_NAME);
        _builder.append(" = new ");
        JvmTypeReference _typeRef_1 = SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(ArrayList.class);
        _builder.append(_typeRef_1);
        _builder.append("();");
        _builder.newLineIfNotEmpty();
        _builder.newLine();
        {
          EList<LawNode> _lawNodes = SingleBlangModelInferrer.this.model.getLawNodes();
          for(final LawNode node : _lawNodes) {
            StringConcatenationClient _componentsMethodBody = SingleBlangModelInferrer.this.componentsMethodBody(node, scope);
            _builder.append(_componentsMethodBody);
            _builder.newLineIfNotEmpty();
          }
        }
        _builder.newLine();
        _builder.append("return ");
        _builder.append(SingleBlangModelInferrer.COMPONENTS_LIST_NAME);
        _builder.append(";");
        _builder.newLineIfNotEmpty();
      }
    };
    return _client;
  }
  
  /**
   * Recursively generate the body of the components() method via a dispatch/recursive collection of methods.
   */
  private StringConcatenationClient _componentsMethodBody(final ForLoop forLoop, final BlangScope scope) {
    JvmTypeReference _iteratorType = forLoop.getIteratorType();
    String _varName = this.varName(forLoop);
    final BlangScope.BlangVariable iteratorVariable = new BlangScope.BlangVariable(_iteratorType, _varName, false);
    final BlangScope childScope = scope.child(iteratorVariable);
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append("for (");
        JvmTypeReference _iteratorType = forLoop.getIteratorType();
        _builder.append(_iteratorType);
        _builder.append(" ");
        String _varName = SingleBlangModelInferrer.this.varName(forLoop);
        _builder.append(_varName);
        _builder.append(" : ");
        StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(forLoop.getIteratorRange(), scope, SingleBlangModelInferrer.this._typeReferenceBuilder.typeRef(Iterable.class, forLoop.getIteratorType()));
        _builder.append(_process);
        _builder.append(") {");
        _builder.newLineIfNotEmpty();
        {
          EList<LawNode> _loopBody = forLoop.getLoopBody();
          for(final LawNode node : _loopBody) {
            _builder.append("  ");
            StringConcatenationClient _componentsMethodBody = SingleBlangModelInferrer.this.componentsMethodBody(node, childScope);
            _builder.append(_componentsMethodBody, "  ");
            _builder.newLineIfNotEmpty();
          }
        }
        _builder.append("}");
        _builder.newLine();
      }
    };
    return _client;
  }
  
  private StringConcatenationClient _componentsMethodBody(final LogScaleFactorDeclaration factor, final BlangScope scope) {
    return this.componentsMethodBody(factor, scope, factor.getContents().getDependencies());
  }
  
  private StringConcatenationClient _componentsMethodBody(final IndicatorDeclaration factor, final BlangScope scope) {
    return this.componentsMethodBody(factor, scope, factor.getContents().getDependencies());
  }
  
  private StringConcatenationClient _componentsMethodBody(final InstantiatedDistribution instantiatedDist, final BlangScope scope) {
    return this.componentsMethodBody(instantiatedDist, scope, instantiatedDist.getDependencies());
  }
  
  /**
   * Common behavior extracted for LogScaleFactorDeclaration, IndicatorDeclaration, InstantiatedDistribution:
   * - taking care of dependencies,
   * - creation of a scope restricted to dependencies,
   * - creating and adding factor to collection.
   */
  private StringConcatenationClient componentsMethodBody(final LawNode node, final BlangScope scope, final List<Dependency> dependencies) {
    final List<InitializerDependency> initializerDependencies = StaticUtils.initializerDependencies(dependencies);
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append("{ // Code generated by: ");
        String _expressionText = StaticUtils.expressionText(node);
        _builder.append(_expressionText);
        _builder.newLineIfNotEmpty();
        _builder.append("  ");
        {
          boolean _isEmpty = initializerDependencies.isEmpty();
          boolean _not = (!_isEmpty);
          if (_not) {
            _builder.append("// Required initialization:");
          }
        }
        _builder.newLineIfNotEmpty();
        {
          for(final InitializerDependency dependency : initializerDependencies) {
            _builder.append("  ");
            JvmTypeReference _type = dependency.getType();
            _builder.append(_type, "  ");
            _builder.append(" ");
            String _name = dependency.getName();
            _builder.append(_name, "  ");
            _builder.append(" = ");
            StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(dependency.getInit(), scope, dependency.getType());
            _builder.append(_process, "  ");
            _builder.append(";");
            _builder.newLineIfNotEmpty();
          }
        }
        _builder.append("  ");
        _builder.append("// Construction and addition of the factor/model:");
        _builder.newLine();
        _builder.append("  ");
        _builder.append(SingleBlangModelInferrer.COMPONENTS_LIST_NAME, "  ");
        _builder.append(".add(");
        _builder.newLineIfNotEmpty();
        _builder.append("    ");
        StringConcatenationClient _instantiateFactor = SingleBlangModelInferrer.this.instantiateFactor(node, 
          scope.restrict(dependencies), scope);
        _builder.append(_instantiateFactor, "    ");
        _builder.append(");");
        _builder.newLineIfNotEmpty();
        _builder.append("}");
        _builder.newLine();
      }
    };
    return _client;
  }
  
  /**
   * Note: in the following documentation, "method1" and "method2" refer to globally, uniquely generated method names.
   * Low-level details needed to cover XExpressions are covered in the class XExpressionProcessor.
   */
  private StringConcatenationClient _instantiateFactor(final LogScaleFactorDeclaration logScaleFactor, final BlangScope scope, final BlangScope parentScope) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        StringConcatenationClient _processLogScaleFactor = SingleBlangModelInferrer.this.xExpressions.processLogScaleFactor(logScaleFactor.getContents().getFactorBody(), scope);
        _builder.append(_processLogScaleFactor);
      }
    };
    return _client;
  }
  
  private StringConcatenationClient _instantiateFactor(final IndicatorDeclaration indic, final BlangScope scope, final BlangScope parentScope) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        StringConcatenationClient _processSupportFactor = SingleBlangModelInferrer.this.xExpressions.processSupportFactor(indic.getContents().getFactorBody(), scope);
        _builder.append(_processSupportFactor);
      }
    };
    return _client;
  }
  
  private String _fullyQualifiedName(final BlangDist blangDist) {
    String _xifexpression = null;
    String _package = blangDist.getDistributionType().getPackage();
    boolean _tripleEquals = (_package == null);
    if (_tripleEquals) {
      _xifexpression = "";
    } else {
      String _package_1 = blangDist.getDistributionType().getPackage();
      _xifexpression = (_package_1 + ".");
    }
    final String prefix = _xifexpression;
    String _name = blangDist.getDistributionType().getName();
    return (prefix + _name);
  }
  
  private String _fullyQualifiedName(final JavaDist javaDist) {
    return javaDist.getDistributionType().getIdentifier();
  }
  
  private StringConcatenationClient _instantiateFactor(final InstantiatedDistribution distribution, final BlangScope scope, final BlangScope parentScope) {
    final List<SingleBlangModelInferrer.ConstructorArgument> constructorArguments = this.constructorParameters(distribution.getTypeSpec());
    final Function1<SingleBlangModelInferrer.ConstructorArgument, Boolean> _function = (SingleBlangModelInferrer.ConstructorArgument it) -> {
      boolean _isParam = it.isParam();
      return Boolean.valueOf((!_isParam));
    };
    final int nRandomVariables = IterableExtensions.size(IterableExtensions.<SingleBlangModelInferrer.ConstructorArgument>filter(constructorArguments, _function));
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append("new ");
        String _fullyQualifiedName = SingleBlangModelInferrer.this.fullyQualifiedName(distribution.getTypeSpec());
        _builder.append(_fullyQualifiedName);
        _builder.append("(");
        _builder.newLineIfNotEmpty();
        {
          int _size = constructorArguments.size();
          ExclusiveRange _doubleDotLessThan = new ExclusiveRange(0, _size, true);
          boolean _hasElements = false;
          for(final int index : _doubleDotLessThan) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(", ", "  ");
            }
            {
              boolean _isParam = constructorArguments.get(index).isParam();
              if (_isParam) {
                _builder.append("  ");
                StringConcatenationClient _processSupplier = SingleBlangModelInferrer.this.xExpressions.processSupplier(distribution.getArguments().get((index - nRandomVariables)), scope, constructorArguments.get(index).deboxedType);
                _builder.append(_processSupplier, "  ");
                _builder.newLineIfNotEmpty();
              } else {
                _builder.append("  ");
                StringConcatenationClient _process = SingleBlangModelInferrer.this.xExpressions.process(distribution.getGeneratedVariables().get(index), parentScope, constructorArguments.get(index).deboxedType);
                _builder.append(_process, "  ");
                _builder.newLineIfNotEmpty();
              }
            }
          }
        }
        _builder.append(")");
        _builder.newLine();
      }
    };
    return _client;
  }
  
  /**
   * For technical reasons, we need to look into other model's constructors to find out which variables are param and
   * which are random. The following deal with this aspect.
   */
  private List<SingleBlangModelInferrer.ConstructorArgument> _constructorParameters(final BlangDist distribution) {
    final BlangModel model = distribution.getDistributionType();
    final List<SingleBlangModelInferrer.ConstructorArgument> result = new ArrayList<SingleBlangModelInferrer.ConstructorArgument>();
    final VariableType[] order = { VariableType.RANDOM, VariableType.PARAM };
    for (final VariableType varType : order) {
      final Function1<VariableDeclaration, Boolean> _function = (VariableDeclaration it) -> {
        VariableType _variableType = it.getVariableType();
        return Boolean.valueOf(Objects.equal(_variableType, varType));
      };
      Iterable<VariableDeclaration> _filter = IterableExtensions.<VariableDeclaration>filter(model.getVariableDeclarations(), _function);
      for (final VariableDeclaration variableDecl : _filter) {
        EList<VariableDeclarationComponent> _components = variableDecl.getComponents();
        for (final VariableDeclarationComponent item : _components) {
          {
            final boolean isParam = StaticUtils.isParam(variableDecl.getVariableType());
            JvmTypeReference _type = variableDecl.getType();
            final SingleBlangModelInferrer.ConstructorArgument argument = new SingleBlangModelInferrer.ConstructorArgument(_type, isParam);
            result.add(argument);
          }
        }
      }
    }
    return result;
  }
  
  private List<SingleBlangModelInferrer.ConstructorArgument> _constructorParameters(final JavaDist distribution) {
    try {
      final Class<?> distributionClass = Class.forName(distribution.getDistributionType().getIdentifier());
      Constructor<?> _xblockexpression = null;
      {
        final Constructor<?>[] constructors = distributionClass.getConstructors();
        int _size = ((List<Constructor<?>>)Conversions.doWrapArray(constructors)).size();
        boolean _tripleNotEquals = (_size != 1);
        if (_tripleNotEquals) {
          throw new RuntimeException();
        }
        _xblockexpression = constructors[0];
      }
      final Constructor<?> constructor = _xblockexpression;
      final List<SingleBlangModelInferrer.ConstructorArgument> result = new ArrayList<SingleBlangModelInferrer.ConstructorArgument>();
      Parameter[] _parameters = constructor.getParameters();
      for (final Parameter parameter : _parameters) {
        {
          boolean isParam = false;
          Annotation[] _annotations = parameter.getAnnotations();
          for (final Annotation annotation : _annotations) {
            Class<? extends Annotation> _annotationType = annotation.annotationType();
            boolean _equals = Objects.equal(_annotationType, Param.class);
            if (_equals) {
              isParam = true;
            }
          }
          JvmTypeReference _xifexpression = null;
          if (isParam) {
            _xifexpression = this.extractSupplierType(parameter);
          } else {
            _xifexpression = this._typeReferenceBuilder.typeRef(parameter.getType());
          }
          final JvmTypeReference typeRef = _xifexpression;
          final SingleBlangModelInferrer.ConstructorArgument argument = new SingleBlangModelInferrer.ConstructorArgument(typeRef, isParam);
          result.add(argument);
        }
      }
      return result;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public JvmTypeReference extractSupplierType(final Parameter supplier) {
    boolean _isAssignableFrom = Supplier.class.isAssignableFrom(supplier.getType());
    boolean _not = (!_isAssignableFrom);
    if (_not) {
      throw new RuntimeException();
    }
    Type _parameterizedType = supplier.getParameterizedType();
    final ParameterizedType supplierType = ((ParameterizedType) _parameterizedType);
    final Type[] supplierTypeArgs = supplierType.getActualTypeArguments();
    int _length = supplierTypeArgs.length;
    boolean _notEquals = (_length != 1);
    if (_notEquals) {
      throw new RuntimeException(("Supplier must have exactly one type argument: " + supplierType));
    }
    return this._typeReferenceBuilder.typeRef(supplierTypeArgs[0].getTypeName());
  }
  
  public SingleBlangModelInferrer(final BlangModel model, final JvmDeclaredType output, final JvmTypesBuilder _typeBuilder, final JvmAnnotationReferenceBuilder _annotationTypesBuilder, final JvmTypeReferenceBuilder _typeReferenceBuilder, final IResourceDescriptionsProvider index, final IQualifiedNameConverter qualNameConverter) {
    this.model = model;
    this.output = output;
    this._typeBuilder = _typeBuilder;
    this._annotationTypesBuilder = _annotationTypesBuilder;
    this._typeReferenceBuilder = _typeReferenceBuilder;
    this.index = index;
    XExpressionProcessor _xExpressionProcessor = new XExpressionProcessor(output, _typeBuilder, _typeReferenceBuilder);
    this.xExpressions = _xExpressionProcessor;
    this.qualNameConverter = qualNameConverter;
  }
  
  private final static String FWD_SIM_METHOD_NAME = StaticUtils.uniqueDeclaredMethod(ForwardSimulator.class);
  
  private final static String COMPONENTS_METHOD_NAME = StaticUtils.uniqueDeclaredMethod(Model.class);
  
  public final static String COMPONENTS_LIST_NAME = "components";
  
  public final static String BUILDER_NAME = "Builder";
  
  private final static String BUILDER_METHOD_NAME = StaticUtils.uniqueDeclaredMethod(ModelBuilder.class);
  
  private final static String IS_FROM_CMD_LINE_ARG = "fromCommandLine";
  
  private StringConcatenationClient componentsMethodBody(final LawNode forLoop, final BlangScope scope) {
    if (forLoop instanceof ForLoop) {
      return _componentsMethodBody((ForLoop)forLoop, scope);
    } else if (forLoop instanceof IndicatorDeclaration) {
      return _componentsMethodBody((IndicatorDeclaration)forLoop, scope);
    } else if (forLoop instanceof InstantiatedDistribution) {
      return _componentsMethodBody((InstantiatedDistribution)forLoop, scope);
    } else if (forLoop instanceof LogScaleFactorDeclaration) {
      return _componentsMethodBody((LogScaleFactorDeclaration)forLoop, scope);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(forLoop, scope).toString());
    }
  }
  
  private StringConcatenationClient instantiateFactor(final LawNode indic, final BlangScope scope, final BlangScope parentScope) {
    if (indic instanceof IndicatorDeclaration) {
      return _instantiateFactor((IndicatorDeclaration)indic, scope, parentScope);
    } else if (indic instanceof InstantiatedDistribution) {
      return _instantiateFactor((InstantiatedDistribution)indic, scope, parentScope);
    } else if (indic instanceof LogScaleFactorDeclaration) {
      return _instantiateFactor((LogScaleFactorDeclaration)indic, scope, parentScope);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(indic, scope, parentScope).toString());
    }
  }
  
  private String fullyQualifiedName(final DistributionTypeSpecification blangDist) {
    if (blangDist instanceof BlangDist) {
      return _fullyQualifiedName((BlangDist)blangDist);
    } else if (blangDist instanceof JavaDist) {
      return _fullyQualifiedName((JavaDist)blangDist);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(blangDist).toString());
    }
  }
  
  private List<SingleBlangModelInferrer.ConstructorArgument> constructorParameters(final DistributionTypeSpecification distribution) {
    if (distribution instanceof BlangDist) {
      return _constructorParameters((BlangDist)distribution);
    } else if (distribution instanceof JavaDist) {
      return _constructorParameters((JavaDist)distribution);
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(distribution).toString());
    }
  }
}
