package ca.ubc.stat.blang.jvmmodel;

import blang.core.LogScaleFactor;
import blang.core.SupportFactor;
import ca.ubc.stat.blang.StaticUtils;
import ca.ubc.stat.blang.jvmmodel.BlangScope;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.function.Supplier;
import org.eclipse.emf.common.util.EList;
import org.eclipse.xtend.lib.annotations.Data;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtend2.lib.StringConcatenationClient;
import org.eclipse.xtext.common.types.JvmConstructor;
import org.eclipse.xtext.common.types.JvmDeclaredType;
import org.eclipse.xtext.common.types.JvmField;
import org.eclipse.xtext.common.types.JvmFormalParameter;
import org.eclipse.xtext.common.types.JvmGenericType;
import org.eclipse.xtext.common.types.JvmMember;
import org.eclipse.xtext.common.types.JvmOperation;
import org.eclipse.xtext.common.types.JvmTypeReference;
import org.eclipse.xtext.common.types.JvmVisibility;
import org.eclipse.xtext.xbase.XExpression;
import org.eclipse.xtext.xbase.jvmmodel.JvmTypeReferenceBuilder;
import org.eclipse.xtext.xbase.jvmmodel.JvmTypesBuilder;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.lib.Pure;
import org.eclipse.xtext.xbase.lib.util.ToStringBuilder;

@Data
@SuppressWarnings("all")
public class XExpressionProcessor {
  public static class PhaseTracker {
    private boolean methodGenerationPhase = true;
    
    private boolean methodGenerationPhase() {
      return this.methodGenerationPhase;
    }
    
    private void endMethodGenerationPhase() {
      this.methodGenerationPhase = false;
    }
  }
  
  private final XExpressionProcessor.PhaseTracker phaseTracker = new XExpressionProcessor.PhaseTracker();
  
  private final JvmDeclaredType output;
  
  @Extension
  private final JvmTypesBuilder _typeBuilder;
  
  @Extension
  private final JvmTypeReferenceBuilder _typeReferenceBuilder;
  
  public void endAuxiliaryMethodGenerationPhase() {
    this.phaseTracker.endMethodGenerationPhase();
  }
  
  public StringConcatenationClient process(final XExpression xExpression, final BlangScope scope, final JvmTypeReference xExpressionReturnType) {
    final String xExpressionGeneratedName = this.generatedMethodName(xExpression);
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("Auxiliary method generated to translate:");
    _builder.newLine();
    String _expressionText = StaticUtils.expressionText(xExpression);
    _builder.append(_expressionText);
    final String generatedDoc = _builder.toString();
    boolean _methodGenerationPhase = this.phaseTracker.methodGenerationPhase();
    if (_methodGenerationPhase) {
      EList<JvmMember> _members = this.output.getMembers();
      final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
        it.setVisibility(JvmVisibility.PRIVATE);
        it.setStatic(true);
        this._typeBuilder.setDocumentation(it, generatedDoc);
        Iterable<BlangScope.BlangVariable> _variables = scope.variables();
        for (final BlangScope.BlangVariable variable : _variables) {
          EList<JvmFormalParameter> _parameters = it.getParameters();
          JvmFormalParameter _parameter = this._typeBuilder.toParameter(xExpression, variable.deboxedName, variable.deboxedType);
          this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
        }
        this._typeBuilder.setBody(it, xExpression);
      };
      JvmOperation _method = this._typeBuilder.toMethod(xExpression, xExpressionGeneratedName, xExpressionReturnType, _function);
      this._typeBuilder.<JvmOperation>operator_add(_members, _method);
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("<methodGenerationPhase>");
        }
      };
      return _client;
    }
    StringConcatenationClient _client_1 = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append(xExpressionGeneratedName);
        _builder.append("(");
        {
          Iterable<BlangScope.BlangVariable> _variables = scope.variables();
          boolean _hasElements = false;
          for(final BlangScope.BlangVariable variable : _variables) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(", ", "");
            }
            String _deboxingInvocationString = variable.deboxingInvocationString();
            _builder.append(_deboxingInvocationString);
          }
        }
        _builder.append(")");
      }
    };
    return _client_1;
  }
  
  public StringConcatenationClient processSupplier(final XExpression xExpression, final BlangScope scope, final JvmTypeReference suppliedType) {
    return this.processViaFunctionalInterfaceImplementation(xExpression, scope, this._typeReferenceBuilder.typeRef(Supplier.class, suppliedType), StaticUtils.uniqueDeclaredMethod(Supplier.class), suppliedType);
  }
  
  public StringConcatenationClient processLogScaleFactor(final XExpression xExpression, final BlangScope scope) {
    return this.processViaFunctionalInterfaceImplementation(xExpression, scope, this._typeReferenceBuilder.typeRef(LogScaleFactor.class), StaticUtils.uniqueDeclaredMethod(LogScaleFactor.class), this._typeReferenceBuilder.typeRef("double"));
  }
  
  public StringConcatenationClient processSupportFactor(final XExpression xExpression, final BlangScope scope) {
    StringConcatenationClient _client = new StringConcatenationClient() {
      @Override
      protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
        _builder.append("new ");
        JvmTypeReference _typeRef = XExpressionProcessor.this._typeReferenceBuilder.typeRef(SupportFactor.class);
        _builder.append(_typeRef);
        _builder.append("(");
        StringConcatenationClient _processViaFunctionalInterfaceImplementation = XExpressionProcessor.this.processViaFunctionalInterfaceImplementation(xExpression, scope, XExpressionProcessor.this._typeReferenceBuilder.typeRef(SupportFactor.Support.class), StaticUtils.uniqueDeclaredMethod(SupportFactor.Support.class), XExpressionProcessor.this._typeReferenceBuilder.typeRef("boolean"));
        _builder.append(_processViaFunctionalInterfaceImplementation);
        _builder.append(")");
        _builder.newLineIfNotEmpty();
      }
    };
    return _client;
  }
  
  private StringConcatenationClient processViaFunctionalInterfaceImplementation(final XExpression xExpression, final BlangScope scope, final JvmTypeReference functionalInterfaceTypeRef, final String methodToImplement, final JvmTypeReference methodToImplementReturnType) {
    final String xExpressionGeneratedName = this.generatedMethodName(xExpression);
    final String generatedAuxiliaryClassName = (xExpressionGeneratedName + "_class");
    StringConcatenation _builder = new StringConcatenation();
    _builder.append("Auxiliary method generated to translate:");
    _builder.newLine();
    String _expressionText = StaticUtils.expressionText(xExpression);
    _builder.append(_expressionText);
    final String generatedDoc = _builder.toString();
    if (this.phaseTracker.methodGenerationPhase) {
      EList<JvmMember> _members = this.output.getMembers();
      final Procedure1<JvmOperation> _function = (JvmOperation it) -> {
        it.setVisibility(JvmVisibility.PRIVATE);
        it.setStatic(true);
        this._typeBuilder.setDocumentation(it, generatedDoc);
        Iterable<BlangScope.BlangVariable> _variables = scope.variables();
        for (final BlangScope.BlangVariable variable : _variables) {
          EList<JvmFormalParameter> _parameters = it.getParameters();
          JvmFormalParameter _parameter = this._typeBuilder.toParameter(xExpression, variable.deboxedName, variable.deboxedType);
          this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
        }
        this._typeBuilder.setBody(it, xExpression);
      };
      JvmOperation _method = this._typeBuilder.toMethod(xExpression, xExpressionGeneratedName, methodToImplementReturnType, _function);
      this._typeBuilder.<JvmOperation>operator_add(_members, _method);
      final Procedure1<JvmGenericType> _function_1 = (JvmGenericType it) -> {
        it.setStatic(true);
        it.setVisibility(JvmVisibility.PUBLIC);
      };
      final JvmGenericType implementation = this._typeBuilder.toClass(xExpression, generatedAuxiliaryClassName, _function_1);
      EList<JvmTypeReference> _superTypes = implementation.getSuperTypes();
      this._typeBuilder.<JvmTypeReference>operator_add(_superTypes, functionalInterfaceTypeRef);
      EList<JvmMember> _members_1 = implementation.getMembers();
      final Procedure1<JvmOperation> _function_2 = (JvmOperation it) -> {
        it.setVisibility(JvmVisibility.PUBLIC);
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append("return ");
            _builder.append(xExpressionGeneratedName);
            _builder.append("(");
            {
              Iterable<BlangScope.BlangVariable> _variables = scope.variables();
              boolean _hasElements = false;
              for(final BlangScope.BlangVariable variable : _variables) {
                if (!_hasElements) {
                  _hasElements = true;
                } else {
                  _builder.appendImmediate(", ", "");
                }
                String _deboxingInvocationString = variable.deboxingInvocationString();
                _builder.append(_deboxingInvocationString);
              }
            }
            _builder.append(");");
          }
        };
        this._typeBuilder.setBody(it, _client);
      };
      JvmOperation _method_1 = this._typeBuilder.toMethod(xExpression, methodToImplement, methodToImplementReturnType, _function_2);
      this._typeBuilder.<JvmOperation>operator_add(_members_1, _method_1);
      EList<JvmMember> _members_2 = implementation.getMembers();
      final Procedure1<JvmOperation> _function_3 = (JvmOperation it) -> {
        it.setVisibility(JvmVisibility.PUBLIC);
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            _builder.append("return \"");
            String _escape = StaticUtils.escape(StaticUtils.expressionText(xExpression));
            _builder.append(_escape);
            _builder.append("\";");
          }
        };
        this._typeBuilder.setBody(it, _client);
      };
      JvmOperation _method_2 = this._typeBuilder.toMethod(xExpression, "toString", this._typeReferenceBuilder.typeRef(String.class), _function_3);
      this._typeBuilder.<JvmOperation>operator_add(_members_2, _method_2);
      Iterable<BlangScope.BlangVariable> _variables = scope.variables();
      for (final BlangScope.BlangVariable variable : _variables) {
        EList<JvmMember> _members_3 = implementation.getMembers();
        final Procedure1<JvmField> _function_4 = (JvmField it) -> {
          it.setFinal(true);
        };
        JvmField _field = this._typeBuilder.toField(xExpression, variable.boxedName(), variable.boxedType(this._typeReferenceBuilder), _function_4);
        this._typeBuilder.<JvmField>operator_add(_members_3, _field);
      }
      EList<JvmMember> _members_4 = implementation.getMembers();
      final Procedure1<JvmConstructor> _function_5 = (JvmConstructor it) -> {
        Iterable<BlangScope.BlangVariable> _variables_1 = scope.variables();
        for (final BlangScope.BlangVariable variable_1 : _variables_1) {
          EList<JvmFormalParameter> _parameters = it.getParameters();
          JvmFormalParameter _parameter = this._typeBuilder.toParameter(xExpression, variable_1.boxedName(), variable_1.boxedType(this._typeReferenceBuilder));
          this._typeBuilder.<JvmFormalParameter>operator_add(_parameters, _parameter);
        }
        StringConcatenationClient _client = new StringConcatenationClient() {
          @Override
          protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
            {
              Iterable<BlangScope.BlangVariable> _variables = scope.variables();
              for(final BlangScope.BlangVariable variable : _variables) {
                _builder.append("this.");
                String _boxedName = variable.boxedName();
                _builder.append(_boxedName);
                _builder.append(" = ");
                String _boxedName_1 = variable.boxedName();
                _builder.append(_boxedName_1);
                _builder.append(";");
                _builder.newLineIfNotEmpty();
              }
            }
          }
        };
        this._typeBuilder.setBody(it, _client);
      };
      JvmConstructor _constructor = this._typeBuilder.toConstructor(xExpression, _function_5);
      this._typeBuilder.<JvmConstructor>operator_add(_members_4, _constructor);
      EList<JvmMember> _members_5 = this.output.getMembers();
      this._typeBuilder.<JvmGenericType>operator_add(_members_5, implementation);
      StringConcatenationClient _client = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("<methodGenerationPhase>");
        }
      };
      return _client;
    } else {
      StringConcatenationClient _client_1 = new StringConcatenationClient() {
        @Override
        protected void appendTo(StringConcatenationClient.TargetStringConcatenation _builder) {
          _builder.append("new ");
          _builder.append(generatedAuxiliaryClassName);
          _builder.append("(");
          {
            Iterable<BlangScope.BlangVariable> _variables = scope.variables();
            boolean _hasElements = false;
            for(final BlangScope.BlangVariable variable : _variables) {
              if (!_hasElements) {
                _hasElements = true;
              } else {
                _builder.appendImmediate(", ", "");
              }
              String _boxedName = variable.boxedName();
              _builder.append(_boxedName);
            }
          }
          _builder.append(")");
        }
      };
      return _client_1;
    }
  }
  
  private final Map<Integer, Integer> _generatedIds = new LinkedHashMap<Integer, Integer>();
  
  public String generatedMethodName(final Object object) {
    final int hashCode = object.hashCode();
    boolean _containsKey = this._generatedIds.containsKey(Integer.valueOf(hashCode));
    if (_containsKey) {
      Integer _get = this._generatedIds.get(Integer.valueOf(hashCode));
      String _plus = ("" + _get);
      return StaticUtils.generatedName(_plus);
    }
    final int newSerialId = this._generatedIds.size();
    this._generatedIds.put(Integer.valueOf(hashCode), Integer.valueOf(newSerialId));
    return StaticUtils.generatedName(("" + Integer.valueOf(newSerialId)));
  }
  
  public XExpressionProcessor(final JvmDeclaredType output, final JvmTypesBuilder _typeBuilder, final JvmTypeReferenceBuilder _typeReferenceBuilder) {
    super();
    this.output = output;
    this._typeBuilder = _typeBuilder;
    this._typeReferenceBuilder = _typeReferenceBuilder;
  }
  
  @Override
  @Pure
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((this.phaseTracker== null) ? 0 : this.phaseTracker.hashCode());
    result = prime * result + ((this.output== null) ? 0 : this.output.hashCode());
    result = prime * result + ((this._typeBuilder== null) ? 0 : this._typeBuilder.hashCode());
    result = prime * result + ((this._typeReferenceBuilder== null) ? 0 : this._typeReferenceBuilder.hashCode());
    result = prime * result + ((this._generatedIds== null) ? 0 : this._generatedIds.hashCode());
    return result;
  }
  
  @Override
  @Pure
  public boolean equals(final Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    XExpressionProcessor other = (XExpressionProcessor) obj;
    if (this.phaseTracker == null) {
      if (other.phaseTracker != null)
        return false;
    } else if (!this.phaseTracker.equals(other.phaseTracker))
      return false;
    if (this.output == null) {
      if (other.output != null)
        return false;
    } else if (!this.output.equals(other.output))
      return false;
    if (this._typeBuilder == null) {
      if (other._typeBuilder != null)
        return false;
    } else if (!this._typeBuilder.equals(other._typeBuilder))
      return false;
    if (this._typeReferenceBuilder == null) {
      if (other._typeReferenceBuilder != null)
        return false;
    } else if (!this._typeReferenceBuilder.equals(other._typeReferenceBuilder))
      return false;
    if (this._generatedIds == null) {
      if (other._generatedIds != null)
        return false;
    } else if (!this._generatedIds.equals(other._generatedIds))
      return false;
    return true;
  }
  
  @Override
  @Pure
  public String toString() {
    ToStringBuilder b = new ToStringBuilder(this);
    b.add("phaseTracker", this.phaseTracker);
    b.add("output", this.output);
    b.add("_typeBuilder", this._typeBuilder);
    b.add("_typeReferenceBuilder", this._typeReferenceBuilder);
    b.add("_generatedIds", this._generatedIds);
    return b.toString();
  }
  
  @Pure
  public XExpressionProcessor.PhaseTracker getPhaseTracker() {
    return this.phaseTracker;
  }
  
  @Pure
  public JvmDeclaredType getOutput() {
    return this.output;
  }
  
  @Pure
  public JvmTypesBuilder get_typeBuilder() {
    return this._typeBuilder;
  }
  
  @Pure
  public JvmTypeReferenceBuilder get_typeReferenceBuilder() {
    return this._typeReferenceBuilder;
  }
  
  @Pure
  public Map<Integer, Integer> get_generatedIds() {
    return this._generatedIds;
  }
}
