package ca.ubc.stat.blang.scoping;

import blang.core.BlangExtensions;
import com.google.inject.Inject;
import java.util.ArrayList;
import java.util.List;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.xtext.common.types.JvmType;
import org.eclipse.xtext.common.types.util.TypeReferences;
import org.eclipse.xtext.xbase.scoping.batch.ImplicitlyImportedFeatures;

@SuppressWarnings("all")
public class BlangImplicitlyImportedFeatures extends ImplicitlyImportedFeatures {
  @Inject
  private TypeReferences typeReferences;
  
  /**
   * @return all JvmTypes who's static methods are put on the scope of their first argument type (i.e. extension methods).
   */
  @Override
  public List<JvmType> getExtensionClasses(final Resource context) {
    final List<Class<?>> classes = this.getExtensionClasses();
    final List<JvmType> result = this.getTypes(classes, context);
    result.add(this.typeReferences.findDeclaredType("xlinear.MatrixExtensions", context));
    result.add(this.typeReferences.findDeclaredType("blang.types.ExtensionUtils", context));
    result.add(this.typeReferences.findDeclaredType("blang.distributions.Generators", context));
    return result;
  }
  
  @Override
  public List<Class<?>> getExtensionClasses() {
    final List<Class<?>> result = new ArrayList<Class<?>>();
    result.addAll(super.getExtensionClasses());
    result.add(BlangExtensions.class);
    return result;
  }
  
  @Override
  public List<JvmType> getStaticImportClasses(final Resource context) {
    final List<Class<?>> classes = this.getStaticImportClasses();
    final List<JvmType> result = this.getTypes(classes, context);
    result.add(this.typeReferences.findDeclaredType("xlinear.MatrixOperations", context));
    result.add(this.typeReferences.findDeclaredType("bayonet.math.SpecialFunctions", context));
    result.add(this.typeReferences.findDeclaredType("org.apache.commons.math3.util.CombinatoricsUtils", context));
    result.add(this.typeReferences.findDeclaredType("blang.types.StaticUtils", context));
    return result;
  }
  
  @Override
  public List<Class<?>> getStaticImportClasses() {
    final ArrayList<Class<?>> result = new ArrayList<Class<?>>();
    result.addAll(super.getExtensionClasses());
    result.add(Math.class);
    return result;
  }
}
